//+------------------------------------------------------------------+
//|                                           NearestNeighbor_v1.mq4 |
//|   Prediction of future based on the nearest neighbor in the past |
//|                                             Copyright 2010, gpwr |
//|                                               vlad1004@yahoo.com |
//+------------------------------------------------------------------+
#property copyright "gpwr"


#property indicator_chart_window
#property indicator_buffers 2

//--- future model outputs
#property indicator_color1  Red
#property indicator_style1  STYLE_SOLID
#property indicator_width1  1
//--- past model outputs
#property indicator_color2  Blue
#property indicator_style2  STYLE_SOLID
#property indicator_width2  1

//--- indicator inputs
extern int    Npast   =300; // # of past bars in a pattern
extern int    Nfut    =2;  // # of future bars in a pattern (must be < Npast)
extern double JmaPeriod = 10;
extern double JmaPhase  = 0;
extern ENUM_APPLIED_PRICE JmaPrice = PRICE_OPEN;

//--- indicator buffers
double ynn[],xnn[];

//--- global variables
int    PrevBars,si;
double mx[],sxx[],denx[],open[],sumx,sumxx;
bool   FirstTime;


//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
void init()
{
//--- indicator buffers mapping
   IndicatorBuffers(3);
   SetIndexBuffer(0,ynn);
   SetIndexBuffer(1,xnn);
   SetIndexBuffer(2,open);
   IndicatorDigits(Digits);
   IndicatorShortName("1NN("+Npast+")");
   SetIndexShift(0, Nfut);
//--- initialize global variables
   PrevBars  = Bars-1;
   FirstTime = true;
   si = Npast*(Npast + 1)/2;
   
   return;  
}
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int start()
{
   int limit, counted_bars = IndicatorCounted(); 

   if(counted_bars<0) return(-1);
   if(counted_bars>0) counted_bars--;
         limit = MathMin(Bars-counted_bars,Bars-1);
         for(int i=limit;i>=0;i--) open[i] = iSmooth(iMA(NULL,0,1,0,MODE_SMA,JmaPrice,i),JmaPeriod,JmaPhase,i);
         
   if(counted_bars < 1)
   { 
   limit = Bars-1;   
      for(i=limit;i>=0;i--) 
      {
      xnn[i]    = EMPTY_VALUE; 
      ynn[i]    = EMPTY_VALUE; 
      }
   }   

//--- check for insufficient data and new bar
   int bars=Bars;
   if(bars<Npast+Nfut)
   {
   Print("Error: not enough bars in history!");
   return(0);
   }
   if(PrevBars==bars) return(0);
   PrevBars=bars;


//--- main cycle
//--- Compute correlation sums for current pattern
//--- Current pattern starts at i=bars-Npast and ends at i=bars-1
   double my  = 0.0;
   double syy = 0.0;
   for(i=0;i<Npast;i++)
   {
   double y=open[Bars-1-(bars-Npast+i)];
   my  += y*(i + 1);
   syy += y*y*(i + 1);
   }
   double deny=syy*si-my*my;
   
   if(deny<=0)
   {
   Print("Zero or negative syy*si-my*my = ",deny);
   return(0);
   }
   
   deny=MathSqrt(deny);

//--- Compute correlation sums for past patterns
//--- Past patterns start at k=0 and end at k=bars-Npast-Nfut
   ArrayResize(mx  ,bars-Npast-Nfut+1);
   ArrayResize(sxx ,bars-Npast-Nfut+1);
   ArrayResize(denx,bars-Npast-Nfut+1);
   
   int kstart;
   
   if(FirstTime) kstart=0; else kstart=bars-Npast-Nfut;
   FirstTime=false;
   
   for(int k=kstart;k<=bars-Npast-Nfut;k++)
   {
      if(k == 0)
      {
      mx[0] = 0.0;
      sxx[0]= 0.0;
      sumx  = 0.0;
      sumxx = 0.0;
         for(i=0;i<Npast;i++)
         {
         double x = open[Bars-1-i];
         double xx= x*x;
         mx[0] += x*(i+1);
         sxx[0]+= xx*(i+1);
         sumx  += x;
         sumxx += xx;
         }
      }
      else
      {
      double xnew = open[Bars-1-(k+Npast-1)];
      double xold = open[Bars-1-(k-1)];
      //if(k==bars-Npast-Nfut) Print("inew=",Bars-1-(k+Npast-1)," xnew=",DoubleToStr(xnew,Digits)," iold=",Bars-1-(k+1)," xold=",DoubleToStr(xold,Digits));
      mx[k]  = mx[k-1] - sumx  + xnew*Npast;
      sxx[k] = sxx[k-1]- sumxx + xnew*xnew*Npast;
         
      sumx +=xnew-xold;
      sumxx+=xnew*xnew-xold*xold;
      }
   denx[k]=sxx[k]*si-mx[k]*mx[k];
   }

//--- Compute cross-correlation sums and correlation coefficients and find NN
   double sxy[];
   ArrayResize(sxy,bars-Npast-Nfut+1);
   double b,corrMax=0;
   int knn=0;
   for(k=0;k<=bars-Npast-Nfut;k++)
   {
   //--- Compute sxy
   sxy[k]=0.0;
      for(i=0;i<Npast;i++) sxy[k]+=open[Bars-1-(k+i)]*open[Bars-1-(bars-Npast+i)]*(i+1);

   //--- Compute corr coefficient
      
      if(denx[k]<=0)
      {
      Print("Zero or negative sxx[k]*si-mx[k]*mx[k]. Skipping pattern # ",k);
      continue;
      }
       
   double num=sxy[k]*si-mx[k]*my;
   double corr=num/MathSqrt(denx[k])/deny;
   
      if(corr>corrMax)
      {
      corrMax=corr;
      knn=k;
      b=num/denx[k];
      }
   }
   
   Print("Nearest neighbor starts on ",TimeToStr(Time[Bars-1-knn])," and ends on ",TimeToStr(Time[Bars-1-(knn+Npast-1)]),
         ". Its correlation coefficient with current pattern is ",corrMax);

//--- Compute xnn[] and ynn[] by scaling the nearest neighbor
   double delta=open[0]-b*open[Bars-1-(knn+Npast-1)];
   
   for(i=0;i<Npast+Nfut;i++)
   {
   if(i<=Npast-1) xnn[Npast-i-1]      = b*open[Bars-1-(knn+i)]+delta;
   if(i>=Npast-1) ynn[Npast+Nfut-i-1] = b*open[Bars-1-(knn+i)]+delta;
   }

   return(0);
}
//+------------------------------------------------------------------+


//------------------------------------------------------------------
//
//------------------------------------------------------------------
//
//
//
//
//

double wrk[][10];

#define bsmax  5
#define bsmin  6
#define volty  7
#define vsum   8
#define avolty 9

double iSmooth(double price, double length, double phase, int i, int s=0)
{
   if (length <=1) return(price);
   if (ArrayRange(wrk,0) != Bars) ArrayResize(wrk,Bars);
   
   int r = Bars-i-1;  s *= 10;
      if (r==0) { for(int k=0; k<7; k++) wrk[r][k+s]=price; for(; k<10; k++) wrk[r][k+s]=0; return(price); }

   //
   //
   //
   //
   //
   
      double len1   = MathMax(MathLog(MathSqrt(0.5*(length-1)))/MathLog(2.0)+2.0,0);
      double pow1   = MathMax(len1-2.0,0.5);
      double del1   = price - wrk[r-1][bsmax+s];
      double del2   = price - wrk[r-1][bsmin+s];
      double div    = 1.0/(10.0+10.0*(MathMin(MathMax(length-10,0),100))/100);
      int    forBar = MathMin(r,10);
	
         wrk[r][volty+s] = 0;
               if(MathAbs(del1) > MathAbs(del2)) wrk[r][volty+s] = MathAbs(del1); 
               if(MathAbs(del1) < MathAbs(del2)) wrk[r][volty+s] = MathAbs(del2); 
         wrk[r][vsum+s] =	wrk[r-1][vsum+s] + (wrk[r][volty+s]-wrk[r-forBar][volty+s])*div;
         
         //
         //
         //
         //
         //
   
         wrk[r][avolty+s] = wrk[r-1][avolty+s]+(2.0/(MathMax(4.0*length,30)+1.0))*(wrk[r][vsum+s]-wrk[r-1][avolty+s]);
            if (wrk[r][avolty+s] > 0)
               double dVolty = wrk[r][volty+s]/wrk[r][avolty+s]; else dVolty = 0;   
	               if (dVolty > MathPow(len1,1.0/pow1)) dVolty = MathPow(len1,1.0/pow1);
                  if (dVolty < 1)                      dVolty = 1.0;

      //
      //
      //
      //
      //
	        
   	double pow2 = MathPow(dVolty, pow1);
      double len2 = MathSqrt(0.5*(length-1))*len1;
      double Kv   = MathPow(len2/(len2+1), MathSqrt(pow2));

         if (del1 > 0) wrk[r][bsmax+s] = price; else wrk[r][bsmax+s] = price - Kv*del1;
         if (del2 < 0) wrk[r][bsmin+s] = price; else wrk[r][bsmin+s] = price - Kv*del2;
	
   //
   //
   //
   //
   //
      
      double R     = MathMax(MathMin(phase,100),-100)/100.0 + 1.5;
      double beta  = 0.45*(length-1)/(0.45*(length-1)+2);
      double alpha = MathPow(beta,pow2);

         wrk[r][0+s] = price + alpha*(wrk[r-1][0+s]-price);
         wrk[r][1+s] = (price - wrk[r][0+s])*(1-beta) + beta*wrk[r-1][1+s];
         wrk[r][2+s] = (wrk[r][0+s] + R*wrk[r][1+s]);
         wrk[r][3+s] = (wrk[r][2+s] - wrk[r-1][4+s])*MathPow((1-alpha),2) + MathPow(alpha,2)*wrk[r-1][3+s];
         wrk[r][4+s] = (wrk[r-1][4+s] + wrk[r][3+s]); 

   //
   //
   //
   //
   //

   return(wrk[r][4+s]);
}